<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar  7 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include "<"$comp">_control_task.h"
#include "<"$comp">_activity.h"


/* --- local data ---------------------------------------------------------- */


/* <"[--- genom_${comp}_activity_alloc ----------------------------------]"> */

int
genom_<"$comp">_activity_alloc(struct genom_<"$comp">_cntrlstr *cids)
{
  int id;
  int i;

  /* look for the first free activity */
  id = -1;
  for(i = 1; i<=<"$COMP">_MAX_ACTIVITIES; i++) {
    id = (cids->activities.last + i) % <"$COMP">_MAX_ACTIVITIES;
    if (cids->activities.i[id].any.status == ACT_VOID) break;
  }
  if (id < 0) return -1;

  cids->activities.last = id;
  return id;
}


/* <"[--- genom_${comp}_activity_free -----------------------------------]"> */

void
genom_<"$comp">_activity_free(struct genom_<"$comp">_cntrlstr *cids, int id)
{
  cids->activities.i[id].any.status = ACT_VOID;
}


/* === Send final replies ================================================== */

<'foreach s [$component services] {'>
/* <"[$s name]"> */
static __inline__ void
genom_<"$comp">_<"[$s name]">_activity_report(
  struct genom_<"$comp">_cntrlstr *cids,
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
  int report = a->h.state;
  STATUS s;

  /* send final reply */
  if (a->h.status == ACT_INTER) {
    if (a->h.state == <"[[[$component event] member ok] declarator]">)
      report = <"[[[$component event] member stop] declarator]">;
  }
  if (report == <"[[[$component event] member ok] declarator]">)
    s = csServReplySend(
      cids->csserv, a->h.rid, FINAL_REPLY, OK,
      (char *)&a->out, 0, genom_<"$comp">_<"[$s name]">_encode);
  else
    s = csServReplySend(
      cids->csserv, a->h.rid, FINAL_REPLY,
      <"$COMP">_ERR_FLAG | report, NULL, 0, NULL);
  if (s == ERROR) {
    genom_log_warn(1, "could not send output for service %s", "<"[$s name]">");
    s = csServReplySend(
      cids->csserv, a->h.rid, FINAL_REPLY,
      <"$COMP">_ERR_FLAG |
      <"[[[$component event] member serialization] declarator]">,
      NULL, 0, NULL);
    if (s == ERROR) {
      genom_log_warn(1, "discarding output for service %s", "<"[$s name]">");
    } else
      genom_log_warn(0, "invalid output for service %s", "<"[$s name]">");
  }

  /* cleanup */
  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
  genom_<"$comp">_activity_free(cids, a->h.aid);
}

<'}'>

/* <"[--- genom_${comp}_activity_report ---------------------------------]"> */

void
genom_<"$comp">_activity_report(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
  switch(a->sid) {
<'foreach s [$component services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      genom_<"$comp">_<"[$s name]">_activity_report(
        cids, (struct genom_<"$comp">_<"[$s name]">_activity *)a);
      return;
<'}'>
  }

  assert(!"code not reached");
}


/* === Interrupt incompatible activities =================================== */

<'foreach s [$component services] {'>
/* <"[$s name]"> */
static __inline__ int
genom_<"$comp">_<"[$s name]">_interrupt_reqd(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
<'  if {[llength [$s interrupt]]} {'>
  struct genom_activity *c;
  int delay;
  int id;

  delay = 0;
  for(id = 0; id<<"$COMP">_MAX_ACTIVITIES; id++) {
    c = &cids->activities.i[id].any;
    if (c->aid == a->aid) continue;
    if (c->status == ACT_VOID) continue;

    switch(c->sid) {
<'    foreach c [$s interrupt] {'>
      case <"$COMP">_<"[$c name]">_RQSTID:
<'    }'>
        switch(c->status) {
          case ACT_ETHER:
          case ACT_INTER:
            break;

          case ACT_STOP:
            delay = 1;
            break;

          case ACT_RUN:
            if (!c->interrupted) {
              c->interrupted = 1;
              genom_<"$comp">_schedule_exec(cids);
            }
            delay = 1;
            break;

          case ACT_WAIT:
            c->status = ACT_INTER;
            c->state = <"[[[$component event] member stop] declarator]">;
            genom_<"$comp">_schedule_cntrl(cids);
            break;

          case ACT_VOID: assert(!"code not reached");
        }
        break;
    }
  }
  return delay;
<'  } else {'>
  return 0;
<'  }'>
}

<'}'>

/* <"[--- genom_${comp}_interrupt_reqd ----------------------------------]"> */

int
genom_<"$comp">_interrupt_reqd(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
  /* interrupt incompatible activities */
  switch(a->sid) {
<'foreach s [$component services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      return genom_<"$comp">_<"[$s name]">_interrupt_reqd(cids, a);
<'}'>
  }

  assert(!"code not reached");
  return 0;
}
