<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Fri Mar  2 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include "<"$comp">_c_types.h"
#include "typecopy.h"
#include "<"$comp">_parameters.h"
#include "<"$comp">_control_task.h"


/* --- genom_connect_port_codel -------------------------------------------- */

<"[[$component event] declarator]">
genom_connect_port_codel(const char local[128], const char remote[128])
{
  <"[[$component event] declarator s]"> =
  <"[[[$component event] member no_such_inport] declarator]">;

  /* find the inport */
<'foreach p [$component ports in] {'>
<'  if {"multiple" in [$p kind]} {'>
  if (!strncmp(local, "<"[$p name]">/", sizeof("<"[$p name]">/")-1)) {
    s = genom_<"$comp">_<"[$p name]">_connect(
      local + sizeof("<"[$p name]">/")-1, remote, genom_<"$comp">_cids());
  }
<'  } else {'>
  if (!strcmp(local, "<"[$p name]">")) {
    s = genom_<"$comp">_<"[$p name]">_connect(remote, genom_<"$comp">_cids());
  }
<'  }'>
<'}'>

  return s;
}


/* --- genom_connect_remote_codel ------------------------------------------ */

<"[[$component event] declarator]">
genom_connect_remote_codel(const char local[128], const char remote[128])
{
  <"[[$component event] declarator s]"> =
  <"[[[$component event] member no_such_remote] declarator]">;

  /* find the remote */
<'foreach r [$component remotes] {'>
  if (!strcmp(local, "<"[$r name]">")) {
    s = genom_<"$comp">_<"[$r name]">_connect(remote, genom_<"$comp">_cids());
  }
<'}'>

  return s;
}


<'foreach s [$component services] {'>

/* <"[--- Service [$s name] control codels ------------------------------]"> */

<"[[$component event] declarator]">
genom_<"$comp">_<"[$s name]">_controlcb(
  struct genom_<"$comp">_cntrlstr *cids,
  struct genom_<"$comp">_<"[$s name]">_input *in,
  struct genom_<"$comp">_<"[$s name]">_output *out)
{
  <"[[$component event] declarator]"> s =
     <"[[[$component event] member ok] declarator]">;

  /* check allowance (before/after statements) */
<'foreach other [$s after] {'>
  if (!cids->run_map[<"$COMP">_<"[$other name]">_RQSTID])
    return <"[[[$component event] member disallowed] declarator]">;
<'}'>
<'foreach other [$s before] {'>
  if (cids->run_map[<"$COMP">_<"[$other name]">_RQSTID])
    return <"[[[$component event] member disallowed] declarator]">;
<'}'>

  /* copy inout parameters to output */
<'  foreach p [$s parameters inout] {'>
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference out->[$p name]]">,
    <"[[$p type] pass value in->[$p name]]">);
<'  }'>

  /* call validate codel */
<'  if {[llength [$s validate]]} {'>
<'    # build parameter list
    set plist [list]
    foreach p [[$s validate] parameters] {
      switch -glob -- "[$p src]/[$p dir]" {
        ids/*		{ set m "cids->ids[$p member]" }
        local/in	{ set m "in->[$p name]" }
        local/*		{ set m "out->[$p name]" }
        default		{ error "unsupported codel parameter" }
      }
      switch -- [$p dir] {
	in		{ lappend plist [[$p type] pass value $m] }
	inout - out	{ lappend plist [[$p type] pass reference $m] }
        default		{ error "unsupported direction" }
      }
    }'>
  s = <"[[$s validate] invoke $plist]">;
  genom_log_debug("service <"[$s name]"> validation returned %s",
                  genom_<"$comp">_evname(s));
  if (s != <"[[[$component event] member ok] declarator]">) return s;
<'  }'>

  /* copy ids parameters to ids (attributes) */
<'  foreach p [$s parameters in inout] {'>
<'    if {[$p src] == "ids"} {'>
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference cids->ids[$p member]]">,
    <"[[$p type] pass value in->[$p name]]">);
<'    }'>
<'  }'>
<'  foreach p [$s parameters out inout] {'>
<'    if {[$p src] == "ids"} {'>
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference out->[$p name]]">,
    <"[[$p type] pass value cids->ids[$p member]]">);
<'    }'>
<'  }'>

  /* call simple codels */
<'foreach c [$s codels simple] {'>
<'  # build parameter list
    set plist [list]
    foreach p [$c parameters] {
      switch -glob -- "[$p src]/[$p dir]" {
        ids/*		{ set m "cids->ids[$p member]" }
        local/in	{ set m "in->[$p name][$p member]" }
        local/*		{ set m "out->[$p name][$p member]" }
        default		{ error "unsupported codel parameter" }
      }
      switch -- [$p dir] {
	in		{ lappend plist [[$p type] pass value $m] }
	inout - out	{ lappend plist [[$p type] pass reference $m] }
        default		{ error "unsupported direction" }
      }
    }'>
  s = <"[$c invoke $plist]">;
  genom_log_debug("service %s codel %s returned %s",
                  "<"[$s name]">", "<"[$c name]">", genom_<"$comp">_evname(s));
  switch(s) {
    case <"[[[$component event] member ok] declarator]">:
<'  foreach t [$s throw] {'>
    case <"[$t declarator]">:
<'  }'>
      break;

    default:
      genom_log_warn(
        0, "bad return value %s for codel %s in service %s",
        genom_<"$comp">_evname(s), "<"[$c name]">", "<"[$s name]">");
      s = <"[[[$component event] member bad_transition] declarator]">;
      break;
  }
  if (s != <"[[[$component event] member ok] declarator]">) return s;
<'  }'>

  return s;
}

<'}'>
