<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Feb 29 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#ifndef H_<"$COMP">_CONTROL_TASK
#define H_<"$COMP">_CONTROL_TASK

#include <assert.h>
#include <stdarg.h>

#include "csLib.h"
#include "semLib.h"

#include "<"$comp">_c_types.h"
#include "<"$comp">_parameters.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_portlib.h"
#include "<"$comp">_remotelib.h"
#include "<"$comp">_activity.h"


/* --- internal state ------------------------------------------------------ */

extern const char *genom_instance;

struct genom_<"$comp">_cntrlstr {
  int shutdown;

  char mbox_name[64];
  SERV_ID csserv;
  long taskid;
  SEM_ID sync;

<'if {![catch {$component ids} ids]} {'>
  <"[$ids declarator ids]">;
<'}'>

  struct {
<'foreach p [$component ports] {'>
    struct genom_<"$comp">_<"[$p name]">_port <"[$p name]">;
<'}'>
  } ports;

  struct {
<'foreach r [$component remotes] {'>
    struct genom_<"$comp">_<"[$r name]">_remote <"[$r name]">;
<'}'>
  } remotes;

  struct {
<'foreach t [$component tasks] {'>
    struct {
      long taskid;
      int runnable;
      SEM_ID sync;
      <"[[$component event] declarator status]">;
      <"[[$component event] declarator pact]">;
    } <"[$t name]">;
<'}'>
  } tasks;

  /* activities */
  struct genom_activities activities;
  int run_map[<"$COMP">_NRQSTID];
};


/* control task */
struct genom_<"$comp">_cntrlstr *
	genom_<"$comp">_init(void);
void	genom_<"$comp">_fini(struct genom_<"$comp">_cntrlstr *cids);
void	genom_<"$comp">_schedule_cntrl(struct genom_<"$comp">_cntrlstr *cids);
void	genom_<"$comp">_schedule_exec(struct genom_<"$comp">_cntrlstr *cids);

struct genom_<"$comp">_cntrlstr * genom_<"$comp">_cids(void);

/* exec tasks */
<'foreach t [$component tasks] {'>
void *	genom_<"$comp">_<"[$t name]">_exec_task(void *data);
int	genom_<"$comp">_<"[$t name]">_invoke_perm(
		struct genom_<"$comp">_cntrlstr *);
int	genom_<"$comp">_<"[$t name]">_invoke(
		struct genom_<"$comp">_cntrlstr *, struct genom_activity *);
<'}'>


/* metadata */
int	genom_metadata_<"$comp">_init(void);
void	genom_metadata_<"$comp">_fini(void);
int	genom_metadata_<"$comp">_fetch(const char *comp,
                genom_metadata_component *meta);


/* log functions */
void	genom_log_info(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));
void	genom_log_warn(int h2error, const char *format, ...)
  __attribute__ ((format (printf, 2, 3)));
void	genom_log_debug(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));


/* helper macro returning the name of a component event */
static __inline__ const char *
genom_<"$comp">_evname(<"[[$component event] declarator e]">)
{
  static const char * const emap[] = {
<'foreach m [[$component event] members] {'>
    [<"[$m declarator]">] = "<"[$m name]">",
<'}'>
    "<invalid>"
  };
  return emap[(e>=sizeof(emap)/sizeof(*emap)-1)?sizeof(emap)/sizeof(*emap)-1:e];
}

#endif /* H_<"$COMP">_CONTROL_TASK */
