<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Aug  1 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]
set tname [$task name]

lang c
'>
#include "autoconf/acheader.h"

#include "commonStructLib.h"

#include "<"$comp">_control_task.h"


/* === task codels ========================================================= */
<'
foreach e [$task fsm] {
  set codel [$task fsm $e]
'>

/* state <"$e">  */
static __inline__ <"[[$component event] declarator]">
<"$comp">_codel_<"$tname">_<"$e">(struct genom_<"$comp">_cntrlstr *cids)
{
  <"[[$component event] declarator]"> s;

<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom_tinit_<"[mangle [$p type]]">(<"[[$p type] pass reference [$p name]_]">);
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference [$p name]_]">,
    <"[[$p type] pass value cids->ids[$p member]]">);
<'    }'>
<'  } '>

<'    # build parameter list
    set plist [list]
    foreach p [$codel parameters] {
      switch -glob -- "[$codel kind]/[$p src]" {
        */local		{ error "task codel with service parameter" }
        */remote	{ set m "cids->remotes.[[$p remote] name].data" }
        */port		{ set m "cids->ports.[$p name].handle" }
        async/ids	{ set m [$p name]_ }
        sync/ids	{ set m "cids->ids[$p member]" }
        default		{ error "reached unreachable code" }
      }

      switch  -glob -- "[$p dir]/[$p src]" {
	in/* - */port {
          lappend plist [[$p type] pass value $m]
        }
	default {
	  lappend plist [[$p type] pass reference $m]
	}
      }
    }'>

<'if {[$codel kind] == "async"} {'>
  commonStructGive(cids);
<'}'>
  s = <"[$codel invoke $plist]">;
<'if {[$codel kind] == "async"} {'>
  commonStructTake(cids);

<'  foreach p [$codel parameters ids] {'>
<'    if {[$p dir] != "in"} {'>
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference cids->ids[$p member]]">,
    <"[[$p type] pass value [$p name]_]">);
<'    }'>
  genom_tfini_<"[mangle [$p type]]">(<"[[$p type] pass reference [$p name]_]">);
<'  }'>
<'}'>
  return s;
}
<'}'>


/* <"[--- genom_${comp}_${tname}_invoke_perm ----------------------------]"> */

int
genom_<"$comp">_<"$tname">_invoke_perm(struct genom_<"$comp">_cntrlstr *cids)
{
  <"[[$component event] declarator s]">;

  switch(cids->tasks.<"$tname">.pact) {
    case <"[[[$component event] member ether] declarator]">:
<'  if {[lsearch [$task fsm] sleep] < 0} {'>
    case <"[[[$component event] member sleep] declarator]">:
<'  }'>
      break;

<'foreach e [$task fsm] {'>
    case <"[[[$component event] member $e] declarator]">:
      s = <"$comp">_codel_<"$tname">_<"$e">(cids);
      genom_log_debug("permanent activity in task <"$tname"> yielded %s",
                      genom_<"$comp">_evname(s));
      switch(s) {
<'  foreach y [[$task fsm $e] yield] {'>
        case <"[$y declarator]">:
<'  }'>
          cids->tasks.<"$tname">.pact = s;
          return s != <"[[[$component event] member sleep] declarator]">;

<'  if {[llength [$s throw]]} {'>
<'    foreach t [$task throw] {'>
        case <"[$t declarator]">:
<'    }'>
          cids->tasks.<"$tname">.pact = s;
          return -1;
<'  }'>
        default:
          genom_log_warn(0, "bad transition from %s to %s in task %s"
                         " permanent activity",
                         genom_<"$comp">_evname(cids->tasks.<"$tname">.pact),
                         genom_<"$comp">_evname(s), "<"$tname">");
          cids->tasks.<"$tname">.pact =
            <"[[[$component event] member bad_transition] declarator]">;
          return -1;
      }
      break;

<'  }'>
<'if {![llength [$task fsm]]} {'>
    case <"[[[$component event] member start] declarator]">:
      s = <"[[[$component event] member ether] declarator]">;
      cids->tasks.<"$tname">.pact = s;
      return 0;
<'}'>

    default:
      genom_log_warn(0, "bad state %s in task %s permanent activity",
                     genom_<"$comp">_evname(cids->tasks.<"$tname">.pact),
                     "<"$tname">");
      return -1;
  }
  return 0;
}


/* === services codels ===================================================== */

<'foreach s [$task services] {'>
/* <"[--- Service [$s name] ---------------------------------------------]"> */

<'  foreach e [$s fsm] { set codel [$s fsm $e]'>

/* state <"$e"> */
static __inline__ <"[[$component event] declarator]">
<"$comp">_<"[$s name]">_codel_<"$e">(
  struct genom_<"$comp">_cntrlstr *cids,
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
  <"[[$component event] declarator]"> s;

<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom_tinit_<"[mangle [$p type]]">(<"[[$p type] pass reference [$p name]_]">);
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference [$p name]_]">,
    <"[[$p type] pass value cids->ids[$p member]]">);
<'    }'>
<'  } '>

<'    # build parameter list
    set plist [list]
    foreach p [$codel parameters] {
      switch -glob -- "[$codel kind]/[$p src]/[$p dir]" {
        */local/in	{ set m "a->in.[$p name]" }
        */local/*	{ set m "a->out.[$p name]" }
        */remote/*	{ set m "cids->remotes.[[$p remote] name].data" }
        */port/*	{ set m "cids->ports.[$p name].handle" }
        async/ids/*	{ set m [$p name]_ }
        sync/ids/*	{ set m "cids->ids[$p member]" }
        default		{ error "reached unreachable code" }
      }

      switch  -glob -- "[$p dir]/[$p src]" {
	in/* - */port {
          lappend plist [[$p type] pass value $m]
        }
	default {
	  lappend plist [[$p type] pass reference $m]
	}
      }
    }'>
<'    if {[$codel kind] == "async"} {'>
  commonStructGive(cids);
<'    }'>
  s = <"[$codel invoke $plist]">;
<'    if {[$codel kind] == "async"} {'>
  commonStructTake(cids);

<'      foreach p [$codel parameters ids] {'>
<'        if {[$p dir] != "in"} {'>
  genom_tcopy_<"[mangle [$p type]]">(
    <"[[$p type] pass reference cids->ids[$p member]]">,
    <"[[$p type] pass value [$p name]_]">);
<'        }'>
  genom_tfini_<"[mangle [$p type]]">(<"[[$p type] pass reference [$p name]_]">);
<'      }'>
<'    }'>
  return s;
}
<'  }'>

/* invoke codels according to <"[$s name]"> state */
static __inline__ int
<"$comp">_<"$tname">_<"[$s name]">_invoke(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
  <"[[$component event] declarator s]">;

  genom_log_debug("task %s invoking activity %s", "<"$tname">", "<"[$s name]">");
  if (a->interrupted) {
    a->interrupted = 0;
    a->status = ACT_STOP;
    a->state = <"[[[$component event] member stop] declarator]">;
  }

  switch(a->state) {
<'  if {[lsearch [$s fsm] sleep] < 0} {'>
    case <"[[[$component event] member sleep] declarator]">:
      break;
<'  }'>

<'  if {[lsearch [$s fsm] stop] < 0} {'>
    case <"[[[$component event] member stop] declarator]">:
      a->status = ACT_INTER;
      a->state = <"[[[$component event] member ok] declarator]">;
      genom_<"$comp">_schedule_cntrl(cids);
      return 0;
<'  }'>

<'  foreach e [$s fsm] {'>
    case <"[[[$component event] member $e] declarator]">:
      s = <"$comp">_<"[$s name]">_codel_<"$e">(
        cids, (struct genom_<"$comp">_<"[$s name]">_activity *)a);
      genom_log_debug("service <"[$s name]"> in task <"$tname"> yielded %s",
                      genom_<"$comp">_evname(s));
      switch(s) {
<'    foreach y [[$s fsm $e] yield] {'>
        case <"[$y declarator]">:
<'    }'>
          if (s == <"[[[$component event] member ether] declarator]">) {
            if (a->status == ACT_RUN)
              a->status = ACT_ETHER;
            else
              a->status = ACT_INTER;
            a->state = <"[[[$component event] member ok] declarator]">;
            genom_<"$comp">_schedule_cntrl(cids);
            return 0;
          } else {
            a->state = s;
            return s != <"[[[$component event] member sleep] declarator]">;
          }
          break;

<'    if {[llength [$s throw]]} {'>
<'      foreach t [$s throw] {'>
        case <"[$t declarator]">:
<'      }'>
          if (a->status == ACT_RUN)
            a->status = ACT_ETHER;
          else
            a->status = ACT_INTER;
          a->state = s;
          genom_<"$comp">_schedule_cntrl(cids);
          return 0;
<'    }'>

        default:
          genom_log_warn(0, "bad transition from %s to %s in service %s",
                         genom_<"$comp">_evname(a->state),
                         genom_<"$comp">_evname(s), "<"[$s name]">");
          a->status = ACT_INTER;
          a->state =
            <"[[[$component event] member bad_transition] declarator]">;
          genom_<"$comp">_schedule_cntrl(cids);
          return 0;
      }
      break;

<'  }'>
    default:
      genom_log_warn(0, "bad state %s in service %s",
                     genom_<"$comp">_evname(a->state), "<"[$s name]">");
      a->status = ACT_INTER;
      a->state =
        <"[[[$component event] member bad_transition] declarator]">;
      genom_<"$comp">_schedule_cntrl(cids);
      return 0;
  }
  return 0;
}

<'}'>

/* <"[--- genom_${comp}_${tname}_invoke ---------------------------------]"> */

int
genom_<"$comp">_<"$tname">_invoke(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
  switch(a->sid) {
<'  foreach s [$task services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      return <"$comp">_<"$tname">_<"[$s name]">_invoke(cids, a);
<'  }'>
  }

  return 0;
}
